#!/usr/bin/env python
#
#    backup.py v1.0.5
#
#    Copyright (C) 2006/2007/2008  Michele <o-zone@zerozone.it> Pinassi
#
#    This software is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# 24-Feb-06 Added patch by k0wax <k0wax@mail.ru>
# 09-Mar-06 Various Fixes, added file mask instead of extension, added errorLog to have errors report inside mail
# 06-Sep-07 Some fixes and improvements
# 03-Dec-07 ExcludeLog options
# 09-Jan-08 Start working on Distribuited backup
# 25-Jun-08 Some bug fixed
# 26-Jun-08 Added EXECBEFORE and EXECAFTER options

import os, sys, string, time, popen2
from stat import *
import shutil
import smtplib
from fnmatch import *
from socket import *

class BackupClass:

# === HARDCODED STUFF
    # where to store temporary files (this dir will be created when backup start and removed whn backup finish)
    tmpDir = "/tmp/backup"
    # where i can found main.cfg
    cfgDir = "/var/spool/backup"
    # what SMTP server i have to use for notify
    SMTPServer = "127.0.0.1"

# ========================= DO NOT CHANGE ANYTHING BELOW THIS LINE ============================

    sumFileName = "/summary.log"

    backupDirs = []
    fileDB = []
    excludeFiles = []
    excludeLog = []
        
    backupDir = ""
    backupName = "backup-"+time.strftime("%d%m%Y%H%M") # Default backup name
    backupMode = 0 # 0=NONE 1=INCREMENTAL 2=FULL
    backupDaily = 1 # INCREMENTAL
    backupWeekly = 2 # FULL
    backupMonthly = 0 # NONE
    
    weeklyDay = "6" # Giorno della settimana per il backup settimanale, in cui Lunedi = 0 e Domenica = 6
    monthlyDay = "1" # Giorno del mese per il backup mensile
    
    notifyMail = []
    notifyFrom = ""
    isErrorState = 0

    # Options for distribuited backup monitoring
    clientVersion = "1.0.6"
    clientId = ""
    serverHost = False
    serverPort = ""
    # -
    # Other options
    execBeforeBackup = None
    execAfterBackup = None

    errorLog = []
    errorsIgnored = 0

    filesCopied = 0
    filesChecked = 0
    filesSkipped = 0
    filesRemoved = 0
    filesAdded = 0
    totalDirectories = 0

    def loadConfig(self):
	try:
	    cfgFile = open(self.cfgDir+"/main.cfg")
	except Exception,e:
	    print "Error %s"%e
	    return False
	line = string.strip(cfgFile.readline())
        while line != "":
    	    if(line[0] != "#"):
		try:
		    (com,arg) = string.split(line,' ',1)
	        except Exception,e:
	    	    self.errorLog.append("Error %s"%e)
		else:
		    if(com == "TARGETDIR"):
			self.backupDirs.append(arg)
	    	    elif(com == "EXCLUDEEXT"):
			self.excludeFiles.append("*"+arg)			
			self.errorLog.append("WARNING: EXCLUDEEXT %s is deprecated. Please use EXCLUDE *%s instead !"%(arg,arg))
	    	    elif(com == "EXCLUDELOG"):
			self.excludeLog.append(arg)			
	    	    elif(com == "EXCLUDE"):
			self.excludeFiles.append(arg)			
	    	    elif(com == "NOTIFYMAIL"):
			self.notifyMail.append(arg)
		    elif(com == "BACKUPDIR"):
			self.backupDir = arg
		    elif(com == "BACKUPNAME"):
			self.backupName = time.strftime(arg)
		    elif(com == "BACKUPDAILY"):
			if(arg == "INCREMENTAL"):
			    self.backupDaily = 1
			elif(arg == "FULL"):
			    self.backupDaily = 2
			else:
			    self.backupDaily = 0
		    elif(com == "WEEKLY"):
			self.weeklyDay = int(arg)
		    elif(com == "MONTHLY"):
			self.monthlyDay = int(arg)
		    elif(com == "BACKUPWEEKLY"):
			if(arg == "INCREMENTAL"):
			    self.backupWeekly = 1
			elif(arg == "FULL"):
			    self.backupWeekly = 2
			else:
			    self.backupWeekly = 0
		    elif(com == "BACKUPMONTHLY"):
			if(arg == "INCREMENTAL"):
			    self.backupMonthly = 1
			elif(arg == "FULL"):
			    self.backupMonthly = 2
			else:
			    self.backupMonthly = 0
		    elif(com == "NOTIFYFROM"):
			self.notifyFrom = arg
		    # Options for distribuited backup monitoring
		    elif(com == "UNIQUECID"):
			self.clientId = arg
		    elif(com == "SERVERHOST"):
			self.serverHost = arg
		    elif(com == "SERVERPORT"):
			self.serverPort = int(arg)
		    elif(com == "EXECBEFORE"):
			self.execBeforeBackup = arg
		    elif(com == "EXECAFTER"):
			self.execAfterBackup = arg
		    else:
			self.errorLog.append("WARNING: Unknown command %s %s"%(com,arg))
			
	    line = string.strip(cfgFile.readline())
	cfgFile.close()
	return True
    
    def loadFileDB(self):
	try:
	    cfgFile = open(self.cfgDir+"/file.db")
	except Exception,e:
	    self.errorLog.append("Error %s"%e)
	    return False
	line = string.strip(cfgFile.readline())
	while line != "":
	    if(line[0] != "#"):
		try:
		    file = string.split(line,'|',2)
	        except Exception,e:
	    	    self.errorLog.append("Error %s"%e)
	        else:
		    file.append(0)
	    	    self.fileDB.append(file)
	    line = string.strip(cfgFile.readline())
	cfgFile.close()
	return True

    def saveFileDB(self):
	try:
	    cfgFile = open(self.cfgDir+"/file.db","w")
	except Exception,e:
	    self.errorLog.append("Error %s"%e)
	    return
	for dbEntry in self.fileDB:
	    cfgFile.write("%s|%s|%s\n"%(dbEntry[0],dbEntry[1],dbEntry[2]))
	cfgFile.close()

    def fileIsChanged(self,filePath,fileStat):
	if self.backupMode == 2: # Se backup completo...
	    return True
	fileSize = fileStat[ST_SIZE]
	fileChanged = fileStat[ST_MTIME]
	for dbEntry in self.fileDB:
	    if(dbEntry[0] == filePath):
		# Analizza il file se  cambiato
	        if(fileSize != int(dbEntry[1])):
		    return True
	        if(fileChanged != int(dbEntry[2])):
		    return True
		dbEntry[3] = 1
		return False # Ok, non copiare
	return True
    
    def updateFileEntry(self,filePath,fileStat):
	fileSize = fileStat[ST_SIZE]
	fileChanged = fileStat[ST_MTIME]
	for dbEntry in self.fileDB:
	    if(dbEntry[0] == filePath):
	    	dbEntry[1] = fileSize
		dbEntry[2] = fileChanged
		dbEntry[3] = 2
		return
	self.fileDB.append((filePath,fileSize,fileChanged,3))
    
    def isAllowed(self, filename):
	for excl in self.excludeFiles:
	    if fnmatch(filename, excl):
		return False
	return True

    def isLog(self, filename):
	for excl in self.excludeLog:
	    if fnmatch(filename, excl):
		return False
	return True
	
    def cbFile(self,filePath,fileStat):
	self.filesChecked = self.filesChecked + 1
	# Controlla che il file sia abilitato
	fileName = os.path.split(filePath)[1]
	if self.isAllowed(fileName):
	    if self.fileIsChanged(filePath,fileStat):
		if self.copyFile(filePath):
		    # Se copia andata a buon fine, aggiorna nel DB
		    self.updateFileEntry(filePath,fileStat)
	else:
	    self.filesSkipped = self.filesSkipped + 1

    def copyFile(self,filePath):
	print "Copying %s..."%filePath,
	try:
	    shutil.copy2(filePath, self.tmpDir+filePath)
	except Exception,e:
	    self.errorLog.append("ERROR(%s)"%e)
	    print "ERROR (%s)"%e
	    return False
	self.filesCopied = self.filesCopied + 1
	print "OK"
	return True

    def walktree(self,top, callback):
	self.totalDirectories = self.totalDirectories + 1
	# Crea la directory nella struttura temporanea
        try:
    	    os.makedirs(self.tmpDir+top)
        except Exception,e:
    	    self.errorLog.append("Directory not created: %s"%e)
	# Analizza la lista dei files
	try:
	    for f in os.listdir(top):
		pathname = os.path.join(top, f)
		try:
		    fileStat = os.stat(pathname)
	        except Exception,e:
		    if(self.isLog(pathname)):
	    		self.errorLog.append("Skipping invalid file %s (%s)" % (pathname,e))
		    else:
		        self.errorsIgnored = self.errorsIgnored + 1
		    continue		
		mode = fileStat[ST_MODE]
		if S_ISDIR(mode):
		    # It's a directory, recurse into it
		    if os.path.islink(pathname):
			if(self.isLog(pathname)):
			    self.errorLog.append("Skipping link %s" % pathname)
			else:
			    self.errorsIgnored = self.errorsIgnored + 1
		    else:
        		self.walktree(pathname, callback)
		elif S_ISREG(mode):
        	    # It's a file, call the callback function
		    callback(pathname,fileStat)
		else:
		    # Unknown file type, print a message
		    if(self.isLog(pathname)):
		        self.errorLog.append('Skipping %s' % pathname)
		    else:
		        self.errorsIgnored = self.errorsIgnored + 1
	except Exception,e:
	    self.errorLog.append("Something goes wrong while walking: %s" % (e))

    def __init__(self):
	print "1) Initializing"
    
	if self.loadConfig() is not True:
	    print "FATAL: No config file found (%s missing ?)"%self.cfgDir+"/main.cfg"
	    sys.exit(-1)
	
	if self.loadFileDB() is not True:
	    print "WARNING: No FileDB found. Maybe first time run ? Force COMPLETE backup mode"
	    self.backupMode = 2
	else:
	    # Now manage config
	    self.backupMode = self.backupDaily
	    (tm_year,tm_mon,tm_mday,tm_hour,tm_min,tm_sec,tm_wday,tm_yday,tm_isdst) = time.localtime()
	    # Today i have to do Montly backup ?
	    if self.monthlyDay == tm_mday:
		self.backupMode = self.backupMontly
		# or weekly ?	    
	    elif self.weeklyDay == tm_wday:
		self.backupMode = self.backupWeekly
    
    def sendReport(self):
	msg = "From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n"%(self.notifyFrom, ", ".join(self.notifyMail),"Backup summary "+time.asctime())
	if self.backupMode == 2:
	    msg = msg + "COMPLETE Backup summary\r\n\r\n"
	elif self.backupMode == 1:
	    msg = msg + "INCREMENTAL Backup summary\r\n\r\n"
	else:
	    msg = msg + "NONE Backup summary\r\n\r\n"

	msg = msg + "Total files: %d\tTotal Directories: %d\r\n"%(self.filesChecked,self.totalDirectories)
	msg = msg + "Changed files: %d\tAdded:%d\tRemoved:%d\tSkipped: %d\r\n"%(self.filesCopied,self.filesAdded,self.filesRemoved,self.filesSkipped)

	if len(self.errorLog) > 0:
	    msg = msg + "\r\nErrors:\r\n"
    	    for err in self.errorLog:
		msg = msg + " " + err + "\n"
	else:
	    msg = msg + "\r\nNO ERRORS\r\n"
	    
	msg = msg + "\r\nKnown problems (ignored): %d\n\r"%(self.errorsIgnored)
	
	msg = msg + "\r\n\r\nHave a nice day.\r\n"
    
	try:
	    server = smtplib.SMTP(self.SMTPServer)
	    server.set_debuglevel(False)
	    server.sendmail(self.notifyFrom, self.notifyMail, msg)
	    server.quit()
	except Exception,e:
	    print "SEND NOTIFY Error: %s"%e
	    print "===================== DUMP OF REPORT ========================\n"
	    print msg
	    print "=============================================================\n"

    def serverSend(self, mySocket, clientId, commandId, args):
	# Invia identificativo, connettiti al server ed invia i dati
	mySocket.send("%s %s %s\n"%(clientId,commandId,args))               
	# Attendi risposta
	try:
	    data = mySocket.recv(1024)                 
	except Exception,e:
	    self.errorLog.append("SEND TO SERVER Error: %s"%e)
	    return False	
	try:
	    (serverId,comCode,comArg) = string.split(data,' ',2)
	except Exception,e:
	    self.errorLog.append("SEND TO SERVER Error: %s"%e)
	    return False	
	# print "SEND:%s ACK:%s\n"%(comCode,commandId)
	if int(comCode) == int(commandId): # Comando andato a buon fine
	    return True
	return False
		
    def serverConnect(self):
	mySocket = socket(AF_INET, SOCK_STREAM)    
	mySocket.settimeout(10.0)
	try:
	    mySocket.connect((self.serverHost, self.serverPort))
	except Exception,e:
	    self.errorLog.append("CONNECTION TO SERVER Error: %s" % (e))
	    return False
	
	if(self.serverSend(mySocket,self.clientId, 100, self.clientVersion) == True):
	    # Invia il riepilogo del BACKUP
	    if(self.serverSend(mySocket,self.clientId, 110, "BACKUPMODE:%d ISERRORSTATE:%d TOTALFILES:%d TOTALDIRS:%d"%(self.backupMode,self.isErrorState,self.filesChecked,self.totalDirectories)) == False):
	        self.errorLog.append("SEND TO SERVER Error: cannot send 110 to server")
	        return False
	    # Invia dettagli sui files del BACKUP
	    if(self.serverSend(mySocket,self.clientId, 120, "FILESCOPY:%d FILESADD:%d FILESREM:%d FILESSKIP:%d"%(self.filesCopied,self.filesAdded,self.filesRemoved,self.filesSkipped)) == False):
	        self.errorLog.append("SEND TO SERVER Error: cannot send 120 to server")
	        return False
	    # Invia log degli ERRORI	    
	    for err in self.errorLog:
		if(self.serverSend(mySocket,self.clientId, 130, err) == False):
	    	    self.errorLog.append("SEND TO SERVER Error: cannot send 130 to server")
	    	    return False
	else:
	    self.errorLog.append("CONNECTION TO SERVER Error")
	    
	return self.serverSend(mySocket,self.clientId, 999, self.clientVersion)
	
    def startBackup(self):
	if self.execBeforeBackup:
	    print "1.1) Executing %s before backup"%(self.execBeforeBackup),
	    p = popen2.Popen4(self.execBeforeBackup)
	    p.wait() # Will wait for process to end
	    print p.fromchild.read(),
	    print "DONE !"
	
	print "2) Analyzing directory tree"
	# Avvia analisi delle directory di cui fare il backup
	for bckDir in self.backupDirs:
	    self.walktree(bckDir,self.cbFile)
	
	print "3) Creating backup summary"
		
	# Apri summary file
	try:
	    sumFile = open(self.tmpDir+self.sumFileName,"w")
	except Exception,e:
	    print "Error %s"%e
	    return

	for dbEntry in self.fileDB:
	    if dbEntry[3] == 0: # DELETED
		print "File %s deleted."%dbEntry[0]
		self.filesRemoved = self.filesRemoved + 1
		sumFile.write("REM %s\n"%dbEntry[0])
	    if dbEntry[3] == 1: # SAME
		pass	    
	    if dbEntry[3] == 2: # CHANGED
		sumFile.write("CHG %s\n"%dbEntry[0])
		print "File %s changed."%dbEntry[0]		
	    if dbEntry[3] == 3: # ADDED
		self.filesAdded = self.filesAdded + 1
		print "File %s added."%dbEntry[0]
		sumFile.write("ADD %s\n"%dbEntry[0])
    
	sumFile.close()
    
	print "4) Creating archive"
    
	if self.filesCopied > 0 and self.backupMode > 0:
	    if self.backupMode == 2:
		backupTGZ = os.path.join(self.backupDir, self.backupName + ".tar.gz");
	    elif self.backupMode == 1:
		backupTGZ = os.path.join(self.backupDir, self.backupName + ".diff.tar.gz");
	    # Comprimi e crea l'archivio
	    retcode = os.spawnlp(os.P_WAIT, 'tar','tar','czf',backupTGZ,self.tmpDir)
	    if retcode != 0:
		self.errorLog.append("FATAL: Cannot create "+backupTGZ+" in "+self.tmpDir)
		self.isErrorState = 1

	if self.execAfterBackup:
	    myExec = self.execAfterBackup.replace("$BACKUPFILE",backupTGZ)
	    print "4.1) Executing %s after backup"%(myExec),
	    p = popen2.Popen4(myExec)
	    p.wait() # Will wait for process to end
	    print p.fromchild.read(),
	    print "DONE !"

	print "======================================="
	if self.backupMode == 2:
	    print "COMPLETE Backup summary"
	elif self.backupMode == 1:
	    print "INCREMENTAL Backup summary"
	else:
	    print "NONE Backup summary"
    
	print "Total files: %d\tTotal Directories: %d"%(self.filesChecked,self.totalDirectories)
	print "Changed files: %d\tAdded:%d\tRemoved:%d\tSkipped: %d"%(self.filesCopied,self.filesAdded,self.filesRemoved,self.filesSkipped)
	if self.isErrorState == 0:
	    # Salva DB dei files
	    self.saveFileDB()
	else:
	    print "WARNING ! Something important goes WRONG: YOU HAVE TO CHECK AND RUN BACKUP AGAIN !"
	    self.errorLog.append("FATAL ERROR - Backup DB not saved. Check and RUN BACKUP AGAIN.")
	
	# Rimuovi directory temporanea
	try:
	    shutil.rmtree(self.tmpDir)
	except Exception,e:
	    print "Exception: %s"%e
	
	# Now, if activated, send report to backup server for distribuited monitoring
	if(self.serverHost):
	    self.serverConnect()
	
	# And finally send report by mail
	self.sendReport()

if __name__=='__main__':
    backup = BackupClass()
    backup.startBackup()
    