#!/usr/bin/env python
#
#    restore.py v1.0.0 ALPHA
#
#    Copyright (C) 2008  Michele <o-zone@zerozone.it> Pinassi
#
#    This software is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# 23-Jan-08 Start restoration

import os, sys, string, time, datetime
from stat import *
import shutil
from fnmatch import *
import getopt
import re

class RestoreClass:

# === HARDCODED STUFF
    # where to store temporary files (this dir will be created when backup start and removed whn backup finish)
    tmpDir = "/tmp/backup"
    # where i can found main.cfg
    cfgDir = "/var/spool/backup"

# ========================= DO NOT CHANGE ANYTHING BELOW THIS LINE ============================

    fileDB = []

    backupDir = ""
    backupName = "backup-"+time.strftime("%d%m%Y%H%M") # Default backup name
    
    notifyMail = []
    notifyFrom = ""
    isErrorState = 0
    
    beVerbose = False
    basePath = "/"
    searchPattern = ".*"
    fileDict = {}

    filesCopied = 0
    filesChecked = 0
    filesSkipped = 0
    filesRemoved = 0
    filesAdded = 0
    totalDirectories = 0

    def printLog(self, log):
	if self.beVerbose:
	    print log

    def loadConfig(self):
	try:
	    cfgFile = open(self.cfgDir+"/main.cfg")
	except Exception,e:
	    print "Error %s"%e
	    return False
	line = string.strip(cfgFile.readline())
        while line != "":
    	    if(line[0] != "#"):
		try:
		    (com,arg) = string.split(line,' ',1)
	        except Exception,e:
	    	    self.errorLog.append("Error %s"%e)
		else:
		    if(com == "BACKUPDIR"):
			self.backupDir = arg
		    elif(com == "BACKUPNAME"):
			self.backupName = time.strftime(arg)
		    else:
			pass
			
	    line = string.strip(cfgFile.readline())
	cfgFile.close()
	return True
    
    def loadBackup(self,top):
    	try:
	    for f in os.listdir(top):
		pathname = os.path.join(top, f)
		try:
		    fileStat = os.stat(pathname)
	        except Exception,e:
	    	    self.errorLog.append("Skipping invalid file %s (%s)" % (pathname,e))
		    continue		
		mode = fileStat[ST_MODE]
		if S_ISREG(mode):
		    # Ok, controlla se puo' essere un backup...
		    # DIFF backup example: backup-050320070440.diff.tar.gz
		    # FULL backup example: backup-050320071637.tar.gz
		    if fnmatch(pathname, "*/backup-*.tar.gz"):
			# Ok,  un backup
			self.loadBackupFile(pathname)

	except Exception,e:
	    print "ERROR: Something goes wrong while loading backups files: %s" % (e)

    def loadBackupFile(self, backupFile):
	self.printLog("   Loading %s"%(backupFile),)
	tarLog = os.popen("tar tzf %s"%backupFile)
	line = string.strip(tarLog.readline())
	while line != "":
	    if fnmatch(line, "*summary.log"):
		# Prendi il basePath
		self.basePath = os.path.dirname(line)
		# Ok, adesso estrai il summary.log
		summaryLog = line
		self.printLog("OK")
		break
	    line = string.strip(tarLog.readline())
	tarLog.close()
	
	if(summaryLog):
	    self.printLog("   Analyzing %s of %s"%(summaryLog,backupFile),)
	    if(self.loadSummaryLog(backupFile, summaryLog)):
		self.printLog("OK")
	    else:
		self.printLog("ERROR")
	else:
	    self.printLog("ERROR")
	
    def loadSummaryLog(self, backupFile, summaryLog):
	# Estrai data/ora del backup
	# FULL backup example: backup-050320071637.tar.gz
	backupDT = re.search(r'.*backup-(?P<datetime>\d+).',backupFile).group('datetime')
	# 050320071637	
	dateTime = datetime.datetime(int(backupDT[4:8]),int(backupDT[2:4]),int(backupDT[0:2]),int(backupDT[8:10]),int(backupDT[10:12]))	
	# Crea la lista de files
	sumLog = os.popen("tar xvzf %s %s -O 2>/dev/null"%(backupFile,summaryLog))
	line = string.strip(sumLog.readline())
	while line != "":
	    fileEntry = []
	    fileEntry.append(dateTime)	    
	    fileEntry.append(backupFile)
	    try:
		sumLine = string.split(line,' ',1)
	        fileEntry.append(sumLine[0]);
	        fileEntry.append(sumLine[1]);
	    except Exception,e:
	        print "Error %s"%e
		return False
	    else:
		self.fileDB.append(fileEntry)
	    line = string.strip(sumLog.readline())
	sumLog.close()	
	return True

    def getFileList(self,pattern):
	fileDict = {}
	# Compile the regular expressions
	try:
	    fileRe = re.compile(pattern)
	except Exception,e:
	    print "ERROR: Invalid regular expression."
	    return False
	# Now searching for the file(s)....
        for fileEntry in self.fileDB:
	    fileBackupDT = fileEntry[0] # DateTime del backup
	    fileArchive = fileEntry[1] # Archivio in cui si trova il file
	    fileCommand = fileEntry[2] # ADD, CHG o REM
	    fileName = fileEntry[3] # Nome e path del file
	    
	    if(fileRe.match(fileName,1)):
		# Aggiungi file
		if fileCommand == ('ADD'):
		    if fileDict.has_key(fileName) == False:
			fileDict[fileName] = (fileArchive, fileBackupDT)
			if self.beVerbose:
			    print "ADD: %s\t(%s)"%(fileName,fileBackupDT)
			
		# File cambiato/aggiornato - Controlla data
		if fileCommand == ('CHG'):
		    if fileDict.has_key(fileName):
			(tempArchive, tempDT) = fileDict[fileName]
			if(tempDT < fileBackupDT): # Se piu vecchio aggiorna
			    fileDict[fileName] = (fileArchive, fileBackupDT)
			    if self.beVerbose:
				print "CHG: %s\t(%s)"%(fileName,fileBackupDT)
		    else:
			# Aggiungi il file
			fileDict[fileName] = (fileArchive, fileBackupDT)
			if self.beVerbose:
			    print "ADD: %s\t(%s)"%(fileName,fileBackupDT)

		# File rimosso
		if fileCommand == ('REM'):
		    if fileDict.has_key(fileName):
			if fileBackupDT > fileDict[fileName][1]: # Se piu vecchio, rimuovi
		    	    del fileDict[fileName]
			    if self.beVerbose:
				print "REM: %s\t(%s)"%(fileName,fileBackupDT)
		
	return fileDict

    def __init__(self):
	self.printLog("1) Initializing...",)
	if self.loadConfig() is not True:
	    print "FATAL: No config file found (%s missing ?)"%self.cfgDir+"/main.cfg"
	    sys.exit(-1)
	self.printLog("OK")
	
	self.printLog("2) Looking in %s and loading backups ..."%(self.backupDir))
	self.loadBackup(self.backupDir)

	
    def startList(self):
	self.printLog("3) Analyzing and parsing files tree for %s..."%(self.searchPattern))
	    
	self.fileDict = self.getFileList(self.searchPattern)
	
	print "\n========= LIST AVAILABLE BACKUPS =========";
	
	for fileKey in self.fileDict.keys():
	    print "%s\t%s"%(fileKey,self.fileDict[fileKey][1])    

	print "==========================================\n";
	    	
    def startRestore(self):
	self.printLog("3) Analyzing and parsing files tree for %s..."%(self.searchPattern))
	    
	self.fileDict = self.getFileList(self.searchPattern)
	
	print "\n====== RESTORE FILES FROM BACKUPS ======";
	
	for fileKey in self.fileDict.keys():
	    print "Restoring %s..."%(fileKey),
	    # Recupera questo file
	    origFile = self.fileDict[fileKey][0]
	    os.popen("tar xvzf %s %s"%(origFile,self.basePath+fileKey))
	    #
	    print "OK"

	print "==========================================\n";
		

	
if __name__=='__main__':
    if len(sys.argv) < 2:
	print "restore.py\n"
	print "Usage:"
	print "%s LIST (pattern) to view available backups"%sys.argv[0]
	print "%s RESTORE (pattern) to restore files from backups"%sys.argv[0]
	print "(pattern) must be a valid regular expression"
    else:
        restore = RestoreClass()
	if len(sys.argv) > 2 and sys.argv[2]:
	    restore.searchPattern = sys.argv[2]
	
	if sys.argv[1] == "LIST":	
	    restore.startList()

	if sys.argv[1] == "RESTORE":	
	    restore.startRestore()
    